# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .turbine import Turbine
from .wake import Wake
from .farm import Farm
import json
import numpy as np


class InputReader():
    """
    InputReader parses json input files into inputs for FLORIS objects.

    InputReader is a helper class which parses json input files and 
    provides an interface to instantiate model objects in FLORIS. This 
    class handles input validation regarding input type, but does not 
    enforce value checking. It is designed to function as a singleton 
    object, but that is not enforced or required.

    Returns:
        InputReader:  An instantiated InputReader object
    """

    def __init__(self):

        self._validObjects = ["turbine", "wake", "farm"]

        self._turbine_properties = {
            "rotor_diameter": float,
            "hub_height": float,
            "blade_count": int,
            "pP": float,
            "pT": float,
            "generator_efficiency": float,
            "power_thrust_table": dict,
            "yaw_angle": float,
            "tilt_angle": float,
            "TSR": float
        }

        self._wake_properties = {
            "velocity_model": str,
            "deflection_model": str,
            "combination_model": str,
            "parameters": dict
        }

        self._farm_properties = {
            "wind_speed": float,
            "wind_direction": float,
            "turbulence_intensity": float,
            "wind_shear": float,
            "wind_veer": float,
            "air_density": float,
            "layout_x": list,
            "layout_y": list
        }

    def _parseJSON(self, filename):
        """
        Opens the input json file and parses the contents into a python 
        dict.

        Args:
            filename:  A string that is the path to the json input file.

        Returns:
            dict:  A dictionary *data* that contains the json input 
            file.
        """
        with open(filename) as jsonfile:
            data = json.load(jsonfile)
        return data

    def _validateJSON(self, json_dict, type_map):
        """
        Verifies that the expected fields exist in the json input file 
        and validates the type of the input data by casting the fields 
        to appropriate values based on the predefined type maps in.

        Args:
            json_dict: Input dictionary with all elements of type str.
            type_map: Predefined type map dictionary for type checking 
                inputs structured as {"property": type}.

        Returns:
            dict: Validated and correctly typed input property 
            dictionary.
        """

        validated = {}

        # validate the object type
        if "type" not in json_dict:
            raise KeyError("'type' key is required")

        if json_dict["type"] not in self._validObjects:
            raise ValueError("'type' must be one of {}".format(
                ", ".join(self._validObjects)))

        validated["type"] = json_dict["type"]

        # validate the description
        if "description" not in json_dict:
            raise KeyError("'description' key is required")

        validated["description"] = json_dict["description"]

        # validate the properties dictionary
        if "properties" not in json_dict:
            raise KeyError("'properties' key is required")
        # check every attribute in the predefined type dictionary for existence
        # and proper type in the given inputs
        propDict = {}
        properties = json_dict["properties"]
        for element in type_map:
            if element not in properties:
                raise KeyError("'{}' is required for object type '{}'".format(
                    element, validated["type"]))

            value, error = self._cast_to_type(
                type_map[element], properties[element])
            if error is not None:
                raise error("'{}' must be of type '{}'".format(
                    element, type_map[element]))

            propDict[element] = value

        validated["properties"] = propDict

        return validated

    def _cast_to_type(self, typecast, value):
        """
        Casts the string input to the type in typecast.

        Args:
            typcast: type - the type class to use on value.
            value: str - the input string to cast to 'typecast'.

        Returns:
            type or None: position 0 - the casted value.
            None or Error: position 1 - the caught error.
        """
        try:
            return typecast(value), None
        except ValueError:
            return None, ValueError

    def _build_turbine(self, json_dict):
        """
        Instantiates a Turbine object from a given input file.

        Args:
            json_dict: Input dictionary describing a turbine model.

        Returns:
            Turbine: An instantiated Turbine object.
        """
        propertyDict = self._validateJSON(json_dict, self._turbine_properties)
        propertyDict["properties"]["yaw_angle"] = propertyDict["properties"]["yaw_angle"]
        propertyDict["properties"]["tilt_angle"] = propertyDict["properties"]["tilt_angle"]
        return Turbine(propertyDict)

    def _build_wake(self, json_dict):
        """
        Instantiates a Wake object from a given input file.

        Args:
            json_dict: dict - Input dictionary describing a wake model.

        Returns:
            Wake: An instantiated Wake object.
        """
        propertyDict = self._validateJSON(json_dict, self._wake_properties)
        return Wake(propertyDict)

    def _build_farm(self, json_dict, turbine, wake):
        """
        Instantiates a Farm object from a given input file.

        Args:
            json_dict: Input dictionary describing a farm model.
            turbine: :py:class:`floris.simulation.turbine.Turbine` 
                instance used in 
                :py:class:`floris.simulation.farm.Farm`.
            wake: :py:class:`floris.simulation.wake.Wake` instance used 
                in :py:class:`floris.simulation.farm.Farm`.

        Returns:
            Farm: An instantiated Farm object.
        """
        propertyDict = self._validateJSON(json_dict, self._farm_properties)
        return Farm(propertyDict, turbine, wake)

    def read(self, input_file=None, input_dict=None):
        """
        Parses main input file and instantiates floris objects.

        Args:
            input_file: A string path to the json input file.

        Returns:
            Farm: An instantiated FLORIS model of wind farm.
        """
        if input_file is not None:
            json_dict = self._parseJSON(input_file)
        elif input_dict is not None:
            json_dict = input_dict.copy()
        else:
            raise ValueError("Input file or dictionary must be provided")

        turbine = self._build_turbine(json_dict["turbine"])
        wake = self._build_wake(json_dict["wake"])
        farm = self._build_farm(json_dict["farm"], turbine, wake)
        return farm



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""
The :py:obj:`floris.tools` package contains the modules used to drive
FLORIS simulations and perform studies in various areas of research and
analysis.

All modules can be imported with

    >>> import floris.tools

The ``__init__.py`` file enables the import of all modules in this
package so any additional modules should be included there.

Examples:
    >>> import floris.tools
    
    >>> dir(floris.tools)
    ['__builtins__', '__cached__', '__doc__', '__file__', '__loader__',
    '__name__', '__package__', '__path__', '__spec__', 'cut_plane',
    'energy_ratio', 'floris_utilities', 'flow_data',
    'layout_functions', 'optimization', 'plotting', 'power_rose',
    'rews', 'sowfa_utilities', 'visualization', 'wind_rose']
"""

from . import cut_plane
from . import energy_ratio
from . import floris_utilities
from . import flow_data
from . import layout_functions
from . import optimization
from . import plotting
from . import power_rose
from . import rews
from . import sowfa_utilities
from . import visualization
from . import wind_rose



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import pickle
from .input_reader import InputReader


class Floris():
    """
    Top-level class that contains a Floris model.

    Floris is the highest level class of the Floris package. Import 
    this class and instantiate it with a path to an input file to 
    create a Floris model. Use the ``farm`` attribute to access other 
    objects within the model.

    Args:
        input_file: A string that is the path to the json input file.
        input_dict: A dictionary as generated from the input_reader.

    Returns:
        Floris: An instantiated Floris object.
    """

    def __init__(self, input_file=None, input_dict=None):
        self.input_reader = InputReader()
        self.input_file = input_file
        self.input_dict = input_dict
        self._farm = []
        self.add_farm(
            input_file=self.input_file,
            input_dict=self.input_dict
        )

    @property
    def farm(self):
        """
        Property of the Floris object that returns the farm(s) 
        contained within the object.

        Returns:
            Farm: A Farm object, or if multiple farms, a list of Farm 
            objects.

        Examples:
            To get the Farm object(s) stored in a Floris object:

            >>> farms = floris.farm()
        """
        if len(self._farm) == 1:
            return self._farm[0]
        else:
            return self._farm

    @farm.setter
    def farm(self, value):
        if not hasattr(self._farm):
            self._farm = value

    def add_farm(self, input_file=None, input_dict=None):
        """
        A method that adds a farm with user-defined input file to the 
        Floris object.

        Returns:
            *None* -- The :py:class:`floris.simulation.floris` object 
            is updated directly.

        Examples:
            To add a farm to a Floris object using a specific input 
            file:

            >>> floris.add_farm(input_file='input.json')

            To add a farm to a Floris object using the stored input 
            file:

            >>> floris.add_farm()
        """
        self._farm.append(self.input_reader.read(input_file=input_file,
                                                 input_dict=input_dict))

    def list_farms(self):
        """
        A method that lists the farms and relevant farm details stored 
        in Floris object.

        Returns:
            *None* -- The farm infomation is printed to the console.

        Examples:
            To list the current farms in Floris object:

            >>> floris.list_farms()
        """
        for num, farm in enumerate(self._farm):
            print('Farm {}'.format(num))
            print(farm)

    def export_pickle(self, pickle_file):
        """
        A method that exports a farm to a pickle file.

        Returns:
            *None* -- Creates a pickle file.

        Examples:
            To export a farm to a pickle file:

            >>> floris.export_pickle('saved_farm.p')
        """
        pickle.dump(self.farm, open(pickle_file, "wb"))

    def import_pickle(self, pickle_file):
        """
        A method that imports a farm from a pickle file.

        Returns:
            *None* - Loads the farm into the 
            :py:class:`floris.simulation.floris.farm` object.

        Examples:
            To load a pickled farm:

            >>> floris.import_pickle('saved_farm.p')
        """
        self.farm = pickle.load(open(pickle_file, "rb"))



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ..utilities import Vec3
from .wake_combination import WakeCombination
from .flow_field import FlowField
from .turbine_map import TurbineMap
import copy
import numpy as np


class Farm():
    """
    Farm is a class containing the objects that make up a FLORIS model.

    Farm is the container class of the FLORIS package. It brings 
    together all of the component objects after input (i.e., Turbine, 
    Wake, FlowField) and packages everything into the appropriate data 
    type. Farm should also be used as an entry point to probe objects 
    for generating output.

    Args:
        instance_dictionary: A dictionary as generated from the 
            input_reader; it should have the following key-value pairs:

            -   **description**: A string containing a description of 
                the wind farm.
            -   **properties**: A dictionary containing the following 
                key-value pairs:

                -   **wind_speed**: A float that is the wind speed at 
                    hub height (m/s).
                -   **wind_direction**: A float that is the wind 
                    direction (deg).
                -   **turbulence_intensity**: A float that is the 
                    turbulence intensity (expressed as a decimal 
                    fraction).
                -   **wind_shear**: A float that is the power law wind 
                    shear exponent.
                -   **wind_veer**: A float that is the vertical change 
                    in wind direction across the rotor.
                -   **air_density**: A float that is the air 
                    density (kg/m^3).
                -   **layout_x**: A list that contains the 
                    x coordinates of the turbines.
                -   **layout_y**: A list that contains the 
                    y coordinates of the turbines.

        turbine: The Turbine object used in Farm.
        wake: The Wake object used in Farm.

    Returns:
        Farm: An instantiated Farm object.
    """

    def __init__(self, instance_dictionary, turbine, wake):
        self.description = instance_dictionary["description"]
        properties = instance_dictionary["properties"]
        layout_x = properties["layout_x"]
        layout_y = properties["layout_y"]
        self.wake = wake

        self.flow_field = FlowField(
            wind_speed=properties["wind_speed"],
            wind_direction=properties["wind_direction"],
            wind_shear=properties["wind_shear"],
            wind_veer=properties["wind_veer"],
            turbulence_intensity=properties["turbulence_intensity"],
            air_density=properties["air_density"],
            turbine_map=TurbineMap(
                layout_x,
                layout_y,
                [copy.deepcopy(turbine) for ii in range(len(layout_x))]),
            wake=wake
        )

    def __str__(self):
        return \
            "Description: {}\n".format(self.description) + \
            "Wake Model: {}\n".format(self.flow_field.wake.velocity_model) + \
            "Deflection Model: {}\n".format(
                self.flow_field.wake.deflection_model)

    def set_wake_model(self, wake_model):
        """
        This method sets the wake model used.

        Args:
            wake_model: A string containing the wake model used to 
                calculate the wake; Valid wake model options are: 
                "curl", "gauss", "jensen", and "multizone".

        Returns:
            *None* -- The wake model and flow field are updated in 
            the :py:obj:`floris.simulation.flow_field` object.

        Examples:
            To set the wake model:

            >>> floris.farm.set_wake_model('curl')
        """

        valid_wake_models = ['curl', 'gauss', 'jensen', 'multizone']
        if wake_model not in valid_wake_models:
            raise Exception("Invalid wake model. Valid options include: {}.".format(
                ", ".join(valid_wake_models)))

        if wake_model == 'jensen':
            self.flow_field.wake.velocity_model = 'jensen'
            self.flow_field.wake.deflection_model = 'jimenez'
        elif wake_model == 'multizone':
            self.flow_field.wake.velocity_model = 'multizone'
            self.flow_field.wake.deflection_model = 'jimenez'
        elif wake_model == 'gauss':
            self.flow_field.wake.velocity_model = 'gauss'
            self.flow_field.wake.deflection_model = 'gauss'
        elif wake_model == 'curl':
            self.flow_field.wake.velocity_model = 'curl'
            self.flow_field.wake.deflection_model = 'curl'

        self.flow_field.reinitialize_flow_field(
            with_resolution=self.flow_field.wake.velocity_model.model_grid_resolution)

    def set_yaw_angles(self, yaw_angles):
        """
        This method sets yaw angles for all turbines and optionally 
        calculates the new wake velocities and updates them in the 
        flow field.

        Args:
            yaw_angles: A single float that sets a constant yaw angle 
                for all turbines or a list of floats that are unique 
                yaw angles for each turbine in degrees.

        Returns:
            *None* -- The turbines are updated directly and the flow 
            field is updated in the 
            :py:obj:`floris.simulation.flow_field` object.

        Examples:
            To set all the yaw angles to one value:

            >>> floris.farm.set_yaw_angles(20.0)

            To set unique yaw angles for the turbines (for example, 
            a 3 turbine array):

            >>> floris.farm.set_yaw_angles([20.0, 10.0, 0.0])
        """
        if isinstance(yaw_angles, float) or isinstance(yaw_angles, int):
            yaw_angles = [yaw_angles] * len(self.turbines)

        for yaw_angle, turbine in zip(yaw_angles, self.turbines):
            turbine.yaw_angle = yaw_angle

    # Getters & Setters

    @property
    def wind_speed(self):
        """
        This property returns the wind speed for the wind farm.

        Returns:
            float: The current wind speed in the wind farm in m/s.

        Examples:
            To get the wind speed for the wind farm:

            >>> wind_speed = floris.farm.wind_speed()
        """
        return self.flow_field.wind_speed

    @property
    def wind_direction(self):
        """
        This property returns the wind direction for the wind farm.

        Returns:
            float: The current wind direction in the wind farm in 
            degrees.

        Examples:
            To get the wind direction for the wind farm:

            >>> wind_direction = floris.farm.wind_direction()
        """
        return self.flow_field.wind_direction

    @property
    def wind_shear(self):
        """
        This property returns the wind shear power law exponent for 
        the wind farm.

        Returns:
            float: The current wind shear power law exponent in the 
            wind farm.

        Examples:
            To get the wind shear for the wind farm:

            >>> wind_shear = floris.farm.wind_shear()
        """
        return self.flow_field.wind_shear

    @property
    def wind_veer(self):
        """
        This property returns the wind veer -- the vertical change in 
        wind direction across the rotor.

        Returns:
            float: The current vertical change in wind direction 
            across the rotor in degrees.

        Examples:
            To get the wind veer for the wind farm:

            >>> wind_veer = floris.farm.wind_veer()
        """
        return self.flow_field.wind_veer

    @property
    def turbulence_intensity(self):
        """
        This property returns the turbulence intensity for the 
        wind farm.

        Returns:
            float: The current turbulence intensity expressed as a 
            decimal fraction.

        Examples:
            To get the turbulence intensity for the wind farm:

            >>> TI = floris.farm.turbulence_intensity()
        """
        return self.flow_field.turbulence_intensity

    @property
    def air_density(self):
        """
        This property returns the air density for the wind farm.

        Returns:
            float: The current air density in kg/m^3.

        Examples:
            To get the air density for the wind farm:

            >>> air_density = floris.farm.air_density()
        """
        return self.flow_field.air_density

    @property
    def turbine_map(self):
        """
        This property returns the turbine map of the 
        :py:obj:`floris.simulation.flow_field` object associated with 
        the wind farm.

        Returns:
            TurbineMap: A :py:obj:`floris.simulation.turbine_map` 
            object that holds turbine information for the farm.

        Examples:
            To get the turbine map for the wind farm:

            >>> turbine_map = floris.farm.turbine_map()
        """
        return self.flow_field.turbine_map

    @property
    def turbines(self):
        """
        This property returns the list of 
        :py:obj:`floris.simulation.turbine` objects contained in the 
        :py:obj:`floris.simulation.turbine_map` object.

        Returns:
            [Turbine]: A list of :py:obj:`floris.simulation.turbine` 
            objects that hold the turbine information for the wind farm.

        Examples:
            To get a list of turbine objects from the wind farm:

            >>> turbines = floris.farm.turbines()
        """
        return self.turbine_map.turbines



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np


class Vec3():
    def __init__(self, x1, x2=None, x3=None, string_format=None):
        """
        Object containing vector information for coordinates.

        Args:
            x1: [float, float, float] or float -- The first argument
            can be a list of the three vector components or simply the
            first component of the vector.
            x2: float (optional) -- The second component of the vector.
            x3: float (optional) -- The third component of the vector.
            string_format: str (optional) -- The string format to use in the
                overloaded __str__ function.
        """
        if isinstance(x1, list):
            self.x1, self.x2, self.x3 = [x for x in x1]
        else:
            self.x1 = x1
            self.x2 = x2
            self.x3 = x3

        if not (type(self.x1) == type(self.x2) and
                type(self.x1) == type(self.x3) and
                type(self.x2) == type(self.x3)):
                target_type = type(self.x1)
                self.x2 = target_type(self.x2)
                self.x3 = target_type(self.x3)

        if string_format is not None:
            self.string_format = string_format
        else:
            if type(self.x1) in [int]:
                self.string_format = "{:8d}"
            elif type(self.x1) in [float, np.float64]:
                self.string_format = "{:8.3f}"

    def rotate_on_x3(self, theta, center_of_rotation=None):
        """
        Rotate about the x3 coordinate axis by a given angle
        and center of rotation.
        The angle theta should be given in degrees.

        Sets the rotated components on this object and returns
        """
        if center_of_rotation is None:
            center_of_rotation = Vec3(0.0, 0.0, 0.0)
        x1offset = self.x1 - center_of_rotation.x1
        x2offset = self.x2 - center_of_rotation.x2
        self.x1prime = x1offset * cosd(theta) - x2offset * sind(
            theta) + center_of_rotation.x1
        self.x2prime = x2offset * cosd(theta) + x1offset * sind(
            theta) + center_of_rotation.x2
        self.x3prime = self.x3

    def __str__(self):
        template_string = "{} {} {}".format(self.string_format,
                                            self.string_format,
                                            self.string_format)
        return template_string.format(self.x1, self.x2, self.x3)

    def __add__(self, arg):

        if type(arg) is Vec3:
            return Vec3(self.x1 + arg.x1, self.x2 + arg.x2, self.x3 + arg.x3)
        else:
            return Vec3(self.x1 + arg, self.x2 + arg, self.x3 + arg)

    def __sub__(self, arg):
        if type(arg) is Vec3:
            return Vec3(self.x1 - arg.x1, self.x2 - arg.x2, self.x3 - arg.x3)
        else:
            return Vec3(self.x1 - arg, self.x2 - arg, self.x3 - arg)

    def __mul__(self, arg):

        if type(arg) is Vec3:
            return Vec3(self.x1 * arg.x1, self.x2 * arg.x2, self.x3 * arg.x3)
        else:
            return Vec3(self.x1 * arg, self.x2 * arg, self.x3 * arg)

    def __truediv__(self, arg):
        if type(arg) is Vec3:
            return Vec3(self.x1 / arg.x1, self.x2 / arg.x2, self.x3 / arg.x3)
        else:
            return Vec3(self.x1 / arg, self.x2 / arg, self.x3 / arg)

    def __eq__(self, arg):
        return self.x1 == arg.x1 \
            and self.x2 == arg.x2 \
            and self.x3 == arg.x3

    def __hash__(self):
        return hash((self.x1, self.x2, self.x3))


class Output():
    def __init__(self, filename):
        """
        Generate output file at specified location.

        Args:
            filename ([str]): write to location
        """
        self.filename = filename
        self.file = open(self.filename, "w")
        self.ln = "\n"

    def write_empty_line(self):
        """
        Add blank line to file.
        """
        self.write_line("")

    def write_line(self, line):
        """
        Add line to file with specified content.

        Args:
            line ([str]): content to add to file
        """
        self.file.write(line + self.ln)

    def end(self):
        """
        Close file.
        """
        self.file.close()


def cosd(angle):
    """
    cosine of an angle with the angle given in degrees

    Returns:
        [float]: cosine of angle in degrees
    """
    return np.cos(np.radians(angle))


def sind(angle):
    """
    sine of an angle with the angle given in degrees

    Returns:
        [float]: sine of angle in degrees
    """
    return np.sin(np.radians(angle))


def tand(angle):
    """
    tangent of an angle with the angle given in degrees

    Returns:
        [float]: tangent of angle in degrees
    """
    return np.tan(np.radians(angle))


def wrap_180(x):
    """
    Wrap an angle to between -180 and 180

    Returns:
        [array]: angles in specified interval
    """
    x = np.where(x <= -180., x + 360., x)
    x = np.where(x > 180., x - 360., x)
    return (x)


def wrap_360(x):
    """
    Wrap an angle to between 0 and 360

    Returns:
        [array]: angles in specified interval
    """
    x = np.where(x < 0., x + 360., x)
    x = np.where(x >= 360., x - 360., x)
    return (x)



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ..utilities import Vec3
from .turbine import Turbine
import numpy as np
import copy


class TurbineMap():
    """
    TurbineMap contains instances of Turbine for the wind farm.

    TurbineMap is container object which maps a Turbine instance to a 
    :py:class:`floris.utilities.Vec3` object. This class also provides 
    some helper methods for sorting and manipulating the turbine layout.

    Args:
        turbine_map_dict: A dictionary mapping of 
            :py:class:`floris.simulation.turbine.Turbine` to 
            :py:class:`floris.utilities.Vec3`; it should have the 
            following form:

            {

                Vec3(): Turbine(),

                Vec3(): Turbine(),

                ...,

                Vec3(): Turbine()

            }

    Returns:
        TurbineMap: An instantiated TurbineMap object.
    """

    def __init__(self, layout_x, layout_y, turbines):
        """
        all are lists
        """
        turbine_dict = {}
        hub_height = turbines[0].hub_height
        coordinates = list(zip(layout_x, layout_y))
        for i, c in enumerate(coordinates):
            turbine_dict[Vec3(c[0], c[1], hub_height)] = turbines[i]
        self._turbine_map_dict = turbine_dict

    def rotated(self, angle, center_of_rotation):
        """
        Rotate the Turbines in TurbineMap by a specific angle.

        Rotate the turbine coordinates by a given angle about a given 
        center of rotation. This function returns a new TurbineMap 
        object whose turbines are rotated. The original TurbineMap is 
        not modified.

        Args:
            angle: The angle, in degrees, of which to rotate the 
                turbines.
            center_of_rotation: The center of rotation. If not supplied 
                the default is ``Vec3(0.0, 0.0, 0.0)``.

        Returns:
            TurbineMap: A rotated TurbineMap.
        """
        layout_x = np.zeros(len(self.coords))
        layout_y = np.zeros(len(self.coords))
        for i, coord in enumerate(self.coords):
            coord.rotate_on_x3(angle, center_of_rotation)
            layout_x[i] = coord.x1prime
            layout_y[i] = coord.x2prime
        return TurbineMap(layout_x, layout_y, self.turbines)

    def sorted_in_x_as_list(self):
        """
        Returns a sorted list of turbine coordinates and turbines from 
        smallest x1 coordinate to largest x1 coordinate.

        Returns:
            [Vec3, Turbine]: A sorted list of turbine coordinates and 
            turbines.
        """
        coords = sorted(self._turbine_map_dict, key=lambda coord: coord.x1)
        return [(c, self._turbine_map_dict[c]) for c in coords]

    @property
    def turbines(self):
        """
        Property that returns list of Turbine objects in wind farm.

        Returns:
            [Turbine]: A list of Turbine objects.
        """
        return [turbine for _, turbine in self.items]

    @property
    def coords(self):
        """
        Property that returns a list of coordinates of the turbines in 
        a wind farm.

        Returns:
            [Vec3]: A list of turbine coordinates.
        """
        return [coord for coord, _ in self.items]

    @property
    def items(self):
        """
        Property that returns a list with dictionary pairs of 
        coordinates and Turbine objects.

        Returns:
            dict_items: List of dictionary key and value pairs of 
            coordinates and Turbine objects.
        """
        return self._turbine_map_dict.items()



"""
Copyright 2017 NREL

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

# This setup file was taken from https://github.com/kennethreitz/setup.py
# accessed on April 3, 2019.

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'FLORIS'
DESCRIPTION = 'A controls-oriented engineering wake model.'
URL = 'https://github.com/NREL/FLORIS'
EMAIL = 'rafael.mudafort@nrel.gov'
AUTHOR = 'NREL National Wind Technology Center'
REQUIRES_PYTHON = '>=3.3.0'
VERSION = '1.1.6'

# What packages are required for this module to be executed?
REQUIRED = [
    # simulation
    'matplotlib',
    'numpy',
    'pytest',
    'scipy',

    # tools
    'h5pyd',
    'pandas',
    'pyproj',
    'seaborn'
]

# What packages are optional?
EXTRAS = {
    'docs': {
        'readthedocs-sphinx-ext==0.5.15',
        'Sphinx==2.0',
        'sphinxcontrib-napoleon==0.7'
    }
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')
        
        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],

    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='Apache-2.0',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np


class WakeCombination():
    """
    The WakeCombination class provides methods for combining the base flow field with the velocity deficits from the wake models.

    Returns:
        WakeCombination: An instantiated WakeCombination object.
    """

    def __init__(self):
        self.model_string = None

    def __str__(self):
        return self.model_string


class FLS(WakeCombination):
    """
    FLS is a subclass of 
    :py:class:`floris.simulation.wake_combination.WakeCombination` 
    which uses freestream linear superposition to combine the base flow 
    field with the wake velocity deficits.
    """

    def __init__(self):
        super().__init__()
        self.model_string = "fls"

    def function(self, u_field, u_wake):
        """
        This method combines the base flow field with the velocity 
        defecits using freestream linear superpostion.

        Args:
            u_field (np.array): The base flow field.
            u_wake (np.array): The wake to add to the rest of the flow 
                field.

        Returns:
            array: A linear combination of the base flow field and the 
            velocity deficits.
        """
        return u_field + u_wake


class SOSFS(WakeCombination):
    """
    SOSFS is a subclass of 
    :py:class:`floris.simulation.wake_combination.WakeCombination` 
    which uses sum of squares freestream superposition to combine the 
    base flow field with the wake velocity deficits.
    """

    def __init__(self):
        super().__init__()
        self.model_string = "sosfs"

    def function(self, u_field, u_wake):
        """
        This method combines the base flow field with the velocity 
        defecits using sum of squares.

        Args:
            u_field (np.array): The base flow field.
            u_wake (np.array): The wake to add to the rest of the flow 
                field.

        Returns:
            array: A sum of squares combination of the base flow field 
            and the velocity deficits.
        """
        return np.hypot(u_wake, u_field)



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np
from ..utilities import Vec3
from ..utilities import cosd, sind, tand
from scipy.interpolate import griddata


class FlowField():
    """
    Object containing flow field information.

    FlowField is at the core of the FLORIS package. This class handles 
    the domain creation and initialization and computes the flow field 
    based on the input wake model and turbine map. It also contains 
    helper functions for quick flow field visualization.

    Args:
        wind_speed: A float that is the wind speed.
        wind_direction: A float that is the wind direction.
        wind_shear: A float that is the wind shear coefficient.
        wind_veer: A float that is the amount of veer across the rotor.
        turbulence_intensity: A float that is the decimal percentage of 
            turbulence.
        wake: A container class :py:class:`floris.simulation.wake` with 
            wake model information used to calculate the flow field.
        wake_combination: A container class 
            :py:class:`floris.simulation.wake_combination` with wake 
            combination information.
        turbine_map: A :py:obj:`floris.simulation.turbine_map` object 
            that holds turbine information.

    Returns:
        FlowField: An instantiated FlowField object.
    """

    def __init__(self,
                 wind_speed,
                 wind_direction,
                 wind_shear,
                 wind_veer,
                 turbulence_intensity,
                 air_density,
                 wake,
                 turbine_map):

        self.reinitialize_flow_field(
            wind_speed=wind_speed,
            wind_direction=wind_direction,
            wind_shear=wind_shear,
            wind_veer=wind_veer,
            turbulence_intensity=turbulence_intensity,
            air_density=air_density,
            wake=wake,
            turbine_map=turbine_map,
            with_resolution=wake.velocity_model.model_grid_resolution
        )

    def _discretize_turbine_domain(self):
        """
        Create grid points at each turbine
        """
        xt = [coord.x1 for coord in self.turbine_map.coords]
        rotor_points = int(
            np.sqrt(self.turbine_map.turbines[0].grid_point_count))
        x_grid = np.zeros((len(xt), rotor_points, rotor_points))
        y_grid = np.zeros((len(xt), rotor_points, rotor_points))
        z_grid = np.zeros((len(xt), rotor_points, rotor_points))

        for i, (coord, turbine) in enumerate(self.turbine_map.items):
            xt = [coord.x1 for coord in self.turbine_map.coords]
            yt = np.linspace(
                coord.x2 - turbine.rotor_radius,
                coord.x2 + turbine.rotor_radius,
                rotor_points
            )
            zt = np.linspace(
                coord.x3 - turbine.rotor_radius,
                coord.x3 + turbine.rotor_radius,
                rotor_points
            )

            for j in range(len(yt)):
                for k in range(len(zt)):
                    x_grid[i, j, k] = xt[i]
                    y_grid[i, j, k] = yt[j]
                    z_grid[i, j, k] = zt[k]

                    xoffset = x_grid[i, j, k] - coord.x1
                    yoffset = y_grid[i, j, k] - coord.x2
                    x_grid[i, j, k] = xoffset * cosd(-1 * self.wind_direction) - \
                        yoffset * sind(-1 * self.wind_direction) + coord.x1
                    y_grid[i, j, k] = yoffset * cosd(-1 * self.wind_direction) + \
                        xoffset * sind(-1 * self.wind_direction) + coord.x2

        return x_grid, y_grid, z_grid

    def _discretize_freestream_domain(self, xmin, xmax, ymin, ymax, zmin, zmax, resolution):
        """
        Generate a structured grid for the entire flow field domain.
        resolution: Vec3
        """
        x = np.linspace(xmin, xmax, int(resolution.x1))
        y = np.linspace(ymin, ymax, int(resolution.x2))
        z = np.linspace(zmin, zmax, int(resolution.x3))
        return np.meshgrid(x, y, z, indexing="ij")

    def _compute_initialized_domain(self, with_resolution=None):
        if with_resolution is not None:
            xmin, xmax, ymin, ymax, zmin, zmax = self.domain_bounds
            self.x, self.y, self.z = self._discretize_freestream_domain(
                xmin, xmax, ymin, ymax, zmin, zmax, with_resolution)
        else:
            self.x, self.y, self.z = self._discretize_turbine_domain()

        self.u_initial = self.wind_speed * \
            (self.z / self.specified_wind_height)**self.wind_shear
        self.v_initial = np.zeros(np.shape(self.u_initial))
        self.w_initial = np.zeros(np.shape(self.u_initial))

        self.u = self.u_initial.copy()
        self.v = self.v_initial.copy()
        self.w = self.w_initial.copy()

    def _compute_turbine_velocity_deficit(self, x, y, z, turbine, coord, deflection, wake, flow_field):
        return self.wake.velocity_function(x, y, z, turbine, coord, deflection, wake, flow_field)

    def _compute_turbine_wake_deflection(self, x, y, turbine, coord, flow_field):
        return self.wake.deflection_function(x, y, turbine, coord, flow_field)

    def _rotated_grid(self, angle, center_of_rotation):
        xoffset = self.x - center_of_rotation.x1
        yoffset = self.y - center_of_rotation.x2
        rotated_x = xoffset * \
            cosd(angle) - yoffset * \
            sind(angle) + center_of_rotation.x1
        rotated_y = xoffset * \
            sind(angle) + yoffset * \
            cosd(angle) + center_of_rotation.x2
        return rotated_x, rotated_y, self.z

    def _rotated_dir(self, angle, center_of_rotation, rotated_map):

        # get new boundaries for the wind farm once rotated
        x_coord = []
        y_coord = []
        for coord in rotated_map.coords:
            x_coord.append(coord.x1)
            y_coord.append(coord.x2)

        if str(self.wake.velocity_model) == 'curl':
            # re-setup the grid for the curl model
            xmin = np.min(x_coord) - 2 * self.max_diameter
            xmax = np.max(x_coord) + 10 * self.max_diameter
            ymin = np.min(y_coord) - 2 * self.max_diameter
            ymax = np.max(y_coord) + 2 * self.max_diameter
            zmin = 0.1
            zmax = 6 * self.specified_wind_height

            # Save these bounds
            self._xmin = xmin
            self._xmax = xmax
            self._ymin = ymin
            self._ymax = ymax
            self._zmin = zmin
            self._zmax = zmax

            resolution = self.wake.velocity_model.model_grid_resolution
            self.x, self.y, self.z = self._discretize_freestream_domain(
                xmin, xmax, ymin, ymax, zmin, zmax, resolution)
            rotated_x, rotated_y, rotated_z = self._rotated_grid(
                0.0, center_of_rotation)
        else:
            rotated_x, rotated_y, rotated_z = self._rotated_grid(
                self.wind_direction, center_of_rotation)

        return rotated_x, rotated_y, rotated_z

    def _calculate_area_overlap(self, wake_velocities, freestream_velocities, turbine):
        """
        compute wake overlap based on the number of points that are not freestream velocity, i.e. affected by the wake
        """
        count = np.sum(freestream_velocities - wake_velocities <= 0.05)
        return (turbine.grid_point_count - count) / turbine.grid_point_count

    # Public methods

    def set_bounds(self, bounds_to_set=None):
        """
        A method that will set the domain bounds for the wake model.

        This method allows a user to customzie the domain bounds for 
        the current wake model being used, unless the wake model is the 
        Curl model, then a predefined domain is specified and used. If 
        the bounds are not specified, then a pre-defined set of bounds 
        will be used. The bounds consist of the minimum and maximum 
        values in the x-, y-, and z-directions.

        Args:
            bounds_to_set: A list of values representing the mininum 
                and maximum values for the domain 
                [xmin, xmax, ymin, ymax, zmin, zmax] 
                (default is *None*).

        Returns:
            *None* -- The flow field is updated directly in the 
            :py:class:`floris.simulation.floris.flow_field` object.
        """

        # For the curl model, bounds are hard coded
        if self.wake.velocity_model.model_string == 'curl':
            coords = self.turbine_map.coords
            x = [coord.x1 for coord in coords]
            y = [coord.x2 for coord in coords]
            eps = 0.1
            self._xmin = min(x) - 2 * self.max_diameter
            self._xmax = max(x) + 10 * self.max_diameter
            self._ymin = min(y) - 2 * self.max_diameter
            self._ymax = max(y) + 2 * self.max_diameter
            self._zmin = 0 + eps
            self._zmax = 6 * self.specified_wind_height

        # Else, if none provided, use a shorter boundary for other models
        elif bounds_to_set is None:
            coords = self.turbine_map.coords
            x = [coord.x1 for coord in coords]
            y = [coord.x2 for coord in coords]
            eps = 0.1
            self._xmin = min(x) - 2 * self.max_diameter
            self._xmax = max(x) + 10 * self.max_diameter
            self._ymin = min(y) - 2 * self.max_diameter
            self._ymax = max(y) + 2 * self.max_diameter
            self._zmin = 0 + eps
            self._zmax = 2 * self.specified_wind_height

        else:  # Set the boundaries
            self._xmin = bounds_to_set[0]
            self._xmax = bounds_to_set[1]
            self._ymin = bounds_to_set[2]
            self._ymax = bounds_to_set[3]
            self._zmin = bounds_to_set[4]
            self._zmax = bounds_to_set[5]

    def reinitialize_flow_field(self,
                                wind_speed=None,
                                wind_direction=None,
                                wind_shear=None,
                                wind_veer=None,
                                turbulence_intensity=None,
                                air_density=None,
                                wake=None,
                                turbine_map=None,
                                with_resolution=None):
        """
        Reiniaitilzies the flow field when a parameter needs to be 
        updated.

        This method allows for changing/updating a variety of flow 
        related parameters. This would typically be used in loops or 
        optimizations where the user is calculating AEP over a wind 
        rose or investigating wind farm performance at different 
        conditions.

        Args:
            wind_speed: A float that is the wind speed (default is 
                *None*).
            wind_direction: A float that is the wind direction (default 
                is *None*).
            wind_shear: A float that is the wind shear coefficient 
                (default is *None*).
            wind_veer: A float that is the amount of veer across the 
                rotor (default is *None*).
            turbulence_intensity: A float that is a decimal percentage 
                of turbulence (default is *None*).
            air_density: A float that is the air density (default is 
                *None*).
            wake: A container class :py:class:`floris.simulation.wake` 
                with wake model information used to calculate the flow 
                field (default is *None*).
            turbine_map: A :py:obj:`floris.simulation.turbine_map` 
                object that holds turbine information (default is 
                *None*).
            with_resolution: A :py:class:`floris.utilities.Vec3` object 
                that defines the flow field resolution at which to 
                calculate the wake (default is *None*).

        Returns:
            *None* -- The flow field is updated directly in the 
            :py:class:`floris.simulation.floris` object.
        """
        # reset the given parameters
        if turbine_map is not None:
            self.turbine_map = turbine_map
        if wind_speed is not None:
            self.wind_speed = wind_speed
        if wind_direction is not None:
            # frame of reference is west
            self.wind_direction = wind_direction - 270
        if wind_shear is not None:
            self.wind_shear = wind_shear
        if wind_veer is not None:
            self.wind_veer = wind_veer
        if turbulence_intensity is not None:
            self.turbulence_intensity = turbulence_intensity
        if air_density is not None:
            self.air_density = air_density
            for turbine in self.turbine_map.turbines:
                turbine.air_density = self.air_density
        if wake is not None:
            self.wake = wake
        if with_resolution is None:
            with_resolution = self.wake.velocity_model.model_grid_resolution

        # initialize derived attributes and constants
        self.max_diameter = max(
            [turbine.rotor_diameter for turbine in self.turbine_map.turbines])
        self.specified_wind_height = self.turbine_map.turbines[0].hub_height

        # Set the domain bounds
        self.set_bounds()

        # reinitialize the flow field
        self._compute_initialized_domain(with_resolution=with_resolution)

        # reinitialize the turbines
        for turbine in self.turbine_map.turbines:
            turbine.reinitialize_turbine(self.turbulence_intensity)

    def calculate_wake(self, no_wake=False):
        """
        Updates the flow field based on turbine activity.

        This method rotates the turbine farm such that the wind 
        direction is coming from 270 degrees. It then loops over the 
        turbines, updating their velocities, calculating the wake 
        deflection/deficit, and combines the wake with the flow field.

        Args:
            no_wake: A bool that when *True* updates the turbine 
                quantities without calculating the wake or adding the 
                wake to the flow field.

        Returns:
            *None* -- The flow field and turbine properties are updated 
            directly in the :py:class:`floris.simulation.floris` object.
        """

        # define the center of rotation with reference to 270 deg
        center_of_rotation = Vec3(0, 0, 0)

        # Rotate the turbines such that they are now in the frame of reference
        # of the wind direction simpifying computing the wakes and wake overlap
        rotated_map = self.turbine_map.rotated(
            self.wind_direction, center_of_rotation)

        # rotate the discrete grid and turbine map
        rotated_x, rotated_y, rotated_z = self._rotated_dir(
            self.wind_direction, center_of_rotation, rotated_map)

        # sort the turbine map
        sorted_map = rotated_map.sorted_in_x_as_list()

        # calculate the velocity deficit and wake deflection on the mesh
        u_wake = np.zeros(np.shape(self.u))
        v_wake = np.zeros(np.shape(self.u))
        w_wake = np.zeros(np.shape(self.u))
        for coord, turbine in sorted_map:

            # update the turbine based on the velocity at its hub
            turbine.update_velocities(
                u_wake, coord, self, rotated_x, rotated_y, rotated_z)

            # get the wake deflecton field
            deflection = self._compute_turbine_wake_deflection(
                rotated_x, rotated_y, turbine, coord, self)

            # get the velocity deficit accounting for the deflection
            turb_u_wake, turb_v_wake, turb_w_wake = self._compute_turbine_velocity_deficit(
                rotated_x, rotated_y, rotated_z, turbine, coord, deflection, self.wake, self)

            # include turbulence model for the gaussian wake model from Porte-Agel
            if self.wake.velocity_model.model_string == 'gauss':

                # compute area overlap of wake on other turbines and update downstream turbine turbulence intensities
                for coord_ti, turbine_ti in sorted_map:

                    if coord_ti.x1 > coord.x1 and np.abs(coord.x2 - coord_ti.x2) < 2*turbine.rotor_diameter:
                        # only assess the effects of the current wake

                        freestream_velocities = turbine_ti.calculate_swept_area_velocities(
                            self.wind_direction,
                            self.u_initial,
                            coord_ti,
                            rotated_x,
                            rotated_y,
                            rotated_z)

                        wake_velocities = turbine_ti.calculate_swept_area_velocities(
                            self.wind_direction,
                            self.u_initial - turb_u_wake,
                            coord_ti,
                            rotated_x,
                            rotated_y,
                            rotated_z)

                        area_overlap = self._calculate_area_overlap(
                            wake_velocities, freestream_velocities, turbine)
                        if area_overlap > 0.0:
                            turbine_ti.turbulence_intensity = turbine_ti.calculate_turbulence_intensity(
                                self.turbulence_intensity,
                                self.wake.velocity_model,
                                coord_ti,
                                coord,
                                turbine
                            )

            # combine this turbine's wake into the full wake field
            if not no_wake:
                # TODO: why not use the wake combination scheme in every component?
                u_wake = self.wake.combination_function(u_wake, turb_u_wake)
                v_wake = (v_wake + turb_v_wake)
                w_wake = (w_wake + turb_w_wake)

        # apply the velocity deficit field to the freestream
        if not no_wake:
            # TODO: are these signs correct?
            self.u = self.u_initial - u_wake
            self.v = self.v_initial + v_wake
            self.w = self.w_initial + w_wake

        # rotate the grid if it is curl
        if self.wake.velocity_model.model_string == 'curl':
            self.x, self.y, self.z = self._rotated_grid(
                -1 * self.wind_direction, center_of_rotation)

    # Getters & Setters
    @property
    def domain_bounds(self):
        """
        Property that returns the bounds of the flow field domain.

        Returns:
            floats: xmin, xmax, ymin, ymax, zmin, zmax

            The mininmum and maxmimum values of the domain in the x, y, and z directions.

        Examples:
            To get the domain bounds:

            >>> xmin, xmax, ymin, ymax, zmin, zmax = 
            ... floris.farm.flow_field.domain_bounds()
        """
        return self._xmin, self._xmax, self._ymin, self._ymax, self._zmin, self._zmax
